	subroutine funder(nctri,ksiv,costval,gradv,iflag)
	    use shared
		use mlef_mod
		implicit none

		integer, intent(in) :: nctri
		double precision, intent(in), dimension(nctri) :: ksiv
		double precision, intent(out) :: costval
		double precision, intent(out), dimension(nctri) :: gradv
		integer, intent(out) :: iflag

        double precision, dimension(3*nctri) :: ksi
        double precision, dimension(3*nctri) :: grad

		! common variables
		double precision :: costu, costv, costp, uval, vval, phival, duval, dvval, dphival

		double precision, dimension(localn) :: localu
        double precision, dimension(localn) :: localv
        double precision, dimension(localn) :: localphi
        double precision, dimension(localn) :: localdu
        double precision, dimension(localn) :: localdv
        double precision, dimension(localn) :: localdphi
        double precision, dimension(NoMembers) :: localgradu
        double precision, dimension(NoMembers) :: localgradv
        double precision, dimension(NoMembers) :: localgradphi
    ! TODO: move into mlef_mod
        double precision, parameter :: MIN_EIGV = 1.d-6

        double precision :: costu0, costv0, costp0

		integer :: i, j, ln, member

	    ksi(1:NoMembers) = ksiv(1:NoMembers)
        ksi(NoMembers+1:2*NoMembers) = ksiv(1:NoMembers)
        ksi(2*NoMembers+1:3*NoMembers) = ksiv(1:NoMembers)

        localu = matmul(localPerturbationU/sqrt(NoMembers-1.),&
                 matmul(localTempEnsEnsU,ksi(1:NoMembers))) + &
                 localMeanU

        localv = matmul(localPerturbationV/sqrt(NoMembers-1.),&
                 matmul(localTempEnsEnsV,ksi(NoMembers+1:2*NoMembers))) + &
                 localMeanV

        localphi = matmul(localPerturbationPhi/sqrt(NoMembers-1.),&
                   matmul(localTempEnsEnsPhi,ksi(2*NoMembers+1:3*NoMembers))) + &
                   localMeanPhi

        costu = 0.
        costv = 0.
        costp = 0.

        do i=1,localn
			call obs_operator(localu(i),localv(i),localphi(i),uval,vval,phival)

			costu = costu &
			        + (uval-localMeanObsU(i)) &
			        * (uval-localMeanObsU(i))/(UObsStddev**2)

			costv = costv &
			        + (vval-localMeanObsV(i)) &
			        * (vval-localMeanObsV(i))/(VObsStddev**2)

			costp = costp &
			        + (phival-localMeanObsPhi(i)) &
			        * (phival-localMeanObsPhi(i))/(PhiObsStddev**2)
	    enddo

        costu = costu + dot_product(ksi(1:NoMembers), &
                        matmul(localAnalysisCovarU,ksi(1:NoMembers)))

        costv = costv + dot_product(ksi(NoMembers+1:2*NoMembers), &
                        matmul(localAnalysisCovarV,ksi(NoMembers+1:2*NoMembers)))

        costp = costp + dot_product(ksi(2*NoMembers+1:3*NoMembers), &
                        matmul(localAnalysisCovarPhi,ksi(2*NoMembers+1:3*NoMembers)))

!        costu0 = 1.0d3
!        costv0 = 1.0d2
!        costp0 = 1.0d4

!         costu0 = 1.0d0
!         costv0 = 1.0d0
!         costp0 = 1.0d0

        costu0 = sqrt(dot_product(localMeanObsU,localMeanObsU))
        costv0 = sqrt(dot_product(localMeanObsV,localMeanObsV))
        costp0 = sqrt(dot_product(localMeanObsPhi,localMeanObsPhi))

        costval = costu/costu0 + costv/costv0 + costp/costp0

        localgradu = 0.
        localgradv = 0.
        localgradphi = 0.

        ! calculate the gradient
        do i=1,localn
            call obs_operator(localu(i),localv(i),localphi(i),uval,vval,phival)
            call obs_operator_adj(localu(i),localv(i),localphi(i),duval,dvval,dphival)

            localdu(i)  = 2.*(uval-localMeanObsU(i))*duval/(UObsStddev**2)
            localdv(i)  = 2.*(vval-localMeanObsV(i))*dvval/(VObsStddev**2)
            localdphi(i) = 2.*(phival-localMeanObsPhi(i))*dphival/(PhiObsStddev**2)
        enddo

        localgradu = matmul(localTempEnsEnsU,&
                     matmul(transpose(localPerturbationU)/sqrt(NoMembers-1.),localdu))

        localgradv = matmul(localTempEnsEnsV,&
                     matmul(transpose(localPerturbationV)/sqrt(NoMembers-1.),localdv))

        localgradphi = matmul(localTempEnsEnsPhi,&
                       matmul(transpose(localPerturbationPhi)/sqrt(NoMembers-1.),localdphi))

        localgradU   = localgradU + 2*matmul(localAnalysisCovarU,ksi(1:NoMembers))
        localgradV   = localgradV + 2*matmul(localAnalysisCovarV,ksi(NoMembers+1:2*NoMembers))
        localgradPhi = localgradPhi + 2*matmul(localAnalysisCovarPhi,ksi(2*NoMembers+1:3*NoMembers))

!        grad(1:NoMembers) = localgradu(1:NoMembers)/costu0
!        grad(NoMembers+1:2*NoMembers) = localgradv(1:NoMembers)/costv0
!        grad(2*NoMembers+1:3*NoMembers) = localgradphi(1:NoMembers)/costp0

        if (minvar .eq. 1) then
            costval = costu/costu0
            gradv(1:NoMembers) = localgradu(1:NoMembers)/costu0
        elseif (minvar .eq. 2) then
            costval = costv/costv0
            gradv(1:NoMembers) = localgradv(1:NoMembers)/costv0
        else
            costval = costp/costp0
            gradv(1:NoMembers) = localgradphi(1:NoMembers)/costp0
        endif
    return
    end

      subroutine funder_costonly(costval,ksiv,nctri)

      implicit none

      double precision, intent(out) :: costval
      double precision, dimension(nctri), intent(in) :: ksiv
      integer, intent(in) :: nctri
      double precision, dimension(nctri) :: fakegrad
      integer :: iflag

      call funder(nctri,ksiv,costval,fakegrad,iflag)

      end

      subroutine funder_gradonly(grad,ksiv,nctri)

      implicit none

      double precision, dimension(nctri), intent(out) :: grad
      double precision, dimension(nctri), intent(in) :: ksiv
      integer, intent(in) :: nctri
      double precision :: fakecostval
      integer :: iflag

      call funder(nctri,ksiv,fakecostval,grad,iflag)

      end
