      subroutine obs_operator(ul,vl,phil,uval,vval,phival)

      use shared
      implicit none

! parameters
      double precision H_MAX, U_MIN, U_MAX, V_START

      parameter (H_MAX = 20000., U_MIN = -5., U_MAX = 5., V_START = 0.25)

! input variables
      double precision, intent(in) :: ul,vl,phil

! output variables
      double precision, intent(out) :: uval, vval, phival

      if (ObsOpNum .eq. 1) then

         ! smooth case

	      uval = ul
	      vval = vl
	      phival = phil

      elseif (ObsOpNum .eq. 2) then

          ! non-smooth case 1

	      uval = ul
	      vval = vl

	      if (phil .gt. H_MAX) then
	          phival = phil*phil/H_MAX
	      else
	          phival = phil
	      endif

      elseif (ObsOpNum .eq. 3) then

          ! non-smooth case 2

	      if (ul .gt. U_MAX) then
	          uval = ul*ul/U_MAX
	      elseif (ul .lt. U_MIN) then
	          uval = ul*ul*ul/U_MIN/U_MIN
	      else
	          uval = ul
	      endif

	      vval = vl

	      if (phil .gt. H_MAX) then
	          phival = phil*phil/H_MAX
	      else
	          phival = phil
	      endif

      elseif (ObsOpNum .eq. 4) then

          ! non-smooth case 3

          if (ul .gt. U_MAX) then
              uval = ul*ul/U_MAX
          elseif (ul .lt. U_MIN) then
              uval = ul*ul*ul/U_MIN/U_MIN
          else
              uval = ul
          endif

          if (vl .ge. 0.0) then
              vval = log(vl + delta)
          else
              vval = log(-vl + delta)
          end if

          if (phil .gt. H_MAX) then
              phival = phil*phil/H_MAX
          else
              phival = phil
          end if

      else
          ! unknown case

          print *,'Unknown ObsOpNum',ObsOpNum

          stop
      endif
      end

      subroutine obs_operator_adj(ul,vl,phil,uadjl,vadjl,phiadjl)

      use shared
      implicit none

! parameters
      double precision H_MAX, U_MIN, U_MAX, V_START

      parameter (H_MAX = 20000., U_MIN = -5., U_MAX = 5, V_START = 0.25)

! input variables
      double precision, intent(in) :: ul,vl,phil

! output variables
      double precision, intent(out) :: uadjl,vadjl,phiadjl

! local variables
      double precision :: uadj1,uadj2,vadj1,vadj2,phiadj1,phiadj2

      if (ObsOpNum .eq. 1) then

         ! smooth case

          uadjl = 1
          vadjl = 1
          phiadjl = 1

      elseif (ObsOpNum .eq. 2) then

          ! non-smooth case 1

          uadjl = 1
          vadjl = 1

          phiadj1 = 1
          phiadj2 = 2*phil/H_MAX

          call smoothing(phil, phiadj1, phiadj2, ObsOpSmoothingKPhi, H_MAX, ObsOpSmoothing, phiadjl)

      elseif (ObsOpNum .eq. 3) then

          ! non-smooth case 2

          uadj1 = 3*ul*ul/U_MIN/U_MIN
          uadj2 = 1.

          call smoothing(ul, uadj1, uadj2, ObsOpSmoothingKU, U_MIN, ObsOpSmoothing, uadjl)

          uadj1 = uadjl
          uadj2 = 2*ul/U_MAX

          call smoothing(ul, uadj1, uadj2, ObsOpSmoothingKU, U_MAX, ObsOpSmoothing, uadjl)

	      vadjl = 1

          phiadj1 = 1
          phiadj2 = 2*phil/H_MAX

          call smoothing(phil, phiadj1, phiadj2, ObsOpSmoothingKPhi, H_MAX, ObsOpSmoothing, phiadjl)

	  elseif (ObsOpNum .eq. 4) then

	      ! non-smooth case 3

          uadj1 = 3*ul*ul/U_MIN/U_MIN
          uadj2 = 1.

          call smoothing(ul, uadj1, uadj2, ObsOpSmoothingKU, U_MIN, ObsOpSmoothing, uadjl)

          uadj1 = uadjl
          uadj2 = 2*ul/U_MAX

          call smoothing(ul, uadj1, uadj2, ObsOpSmoothingKU, U_MAX, ObsOpSmoothing, uadjl)

          if (ObsOpSmoothing .eq. 0) then
		      if (vl .ge. 0.0) then
		          vadjl = 1./(vl+delta)
		      else
		          vadjl = 1./(vl-delta)
		      end if
		  else

              vadj1 = 1./(vl-delta)
              vadj2 = 0.

              call smoothing(vl, vadj1, vadj2, ObsOpSmoothingKV, -V_START, ObsOpSmoothing, vadjl)

              vadj1 = vadjl
              vadj2 = 1./(vl+delta)

              call smoothing(vl, vadj1, vadj2, ObsOpSmoothingKV, V_START, ObsOpSmoothing, vadjl)
		  endif

          phiadj1 = 1
          phiadj2 = 2*phil/H_MAX

          call smoothing(phil, phiadj1, phiadj2, ObsOpSmoothingKPhi, H_MAX, ObsOpSmoothing, phiadjl)

      else
          ! unknown case

          print *,'Unknown ObsOpNum',ObsOpNum

          stop
      endif
      end

      subroutine smoothing(x, y1, y2, k, p, type, yval)

      implicit none

      double precision, intent(in) :: x, y1, y2, k, p
      integer, intent(in) :: type
      double precision, intent(out) :: yval

      if (type .eq. 0) then
          if (x .le. p) then
              yval = y1
          else
              yval = y2
          endif
      elseif (type .eq. 1) then
          yval = 0.5*(y1 + y2 + tanh(k*(p-x))*y1 + tanh(k*(x-p))*y2)
      endif
	  end
