      program swethreedvar

      use shared
      implicit none

! L-BFGS variables
      integer :: ntrav, msave, nxm, nwork

      common /va15dd/ mp, lp, gtol

      integer :: mp, lp
      double precision :: gtol, oldcost

      double precision, allocatable, dimension(:) :: diag, ss, yy, ww
      double precision :: dzs(10), apd

      integer :: iprint(2), point
      logical :: diagco

! LMBM variables
      integer :: na, mcu, nw
      integer :: ipar(7),iout(3)

      double precision, allocatable, dimension(:) :: w
      double precision :: rpar(8)
      double precision :: time,rtim(2)
      integer :: mc, mcinit

! DESCON variables
      integer nexp, iter, irs, fgcnt, lscnt, ibeta, itheta
      integer itert, irstot, fgtot,lstot, stoptest
      double precision gnorm, epsm
      double precision dc, cc

      logical angle, powell

! CG-Descent variables
      double precision, allocatable, dimension(:) :: dtemp, xtemp, gtemp
      integer :: fcnt, gcnt
      external :: funder_costonly, funder_gradonly

! local variables
      double precision, allocatable, dimension(:) :: grad, z, x
      double precision, allocatable, dimension(:,:) :: utmp, vtmp, phitmp
      double precision :: costval, costu, costv, costp, ensfactor
      double precision :: eps

      integer :: i, j, k, kpass, icall, iflag, ifunder, ierr

      character(len=64) :: costFileName, rmseFileName

      call initialize()
      call readFiles()

! initialize L-BFGS variables
      ntrav = 25*nctr
      msave = 9
      nxm = msave*nctr
      nwork = nctr+2*msave

      allocate (diag(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for diag"

      allocate (ss(nxm), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for ss"

      allocate (yy(nxm), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for yy"

      allocate (ww(nwork), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for ww"

! initialize LMBM variables
      na = 4
      mcu = 9
      nw = 1 + 9*nctr + 2*nctr*na + 3*na + 2*nctr*(mcu+1) + 3*(mcu+2)*(mcu+1)/2 + 9*(mcu+1)

      allocate (w(nw), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for w"

! initialize local variables
      allocate (grad(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for grad"

      allocate (z(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for z"

      allocate (x(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for x"

      allocate (utmp(n, m), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for utmp"

      allocate (vtmp(n, m), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for vtmp"

      allocate (phitmp(n, m), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for phitmp"

! initialize CG-descent variables
      allocate (dtemp(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for dtemp"

      allocate (xtemp(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for xtemp"

      allocate (gtemp(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for gtemp"


!      print*,'Now conducting alpha test of gradient correctness'
!      call alphaTest(1.d0)

      u = 0.
      v = 0.
      phi = 0.

      u(   2:n-1, 2:m-1, 1) = uback(   1:n-2, 1:m-2)
      v(   2:n-1, 2:m-1, 1) = vback(   1:n-2, 1:m-2)
      phi( 2:n-1, 2:m-1, 1) = phiback( 1:n-2, 1:m-2)

      call swe_model(u, v, phi, n, m, nt, 1, 1, nt)
      call cost_3dvar(u, v, phi, n, m, nt, 1, uobs, vobs, phiobs, costu, costv, costp, &
          UObsStddev, VObsStddev, PhiObsStddev)

      print *,'Initial cost is:',costval
      print *,'Initial RMSE is:'

      do k=1,nt
          call calculateRmse(k)
      enddo

      print *, 'Now running test case with: (ObsOpNum,OptimizerNum,ObsOpSmoothing):', ObsOpNum, OptimizerNum, ObsOpSmoothing
      print *, '----------------------------------------------------------------------------------------------------------'

      ! set variables used in common
      z = 0

 80   format('Cost-3DVar-',a4,'-Obsop',i1,'-Smoothing',i1,'.txt')
 81   format('RMSE-3DVar-',a4,'-Obsop',i1,'-Smoothing',i1,'.txt')

      if (OptimizerNum .eq. 1) then ! L-BFGS

          write(costFileName,80) 'LMQN', ObsOpNum, ObsOpSmoothing
          write(rmseFileName,81) 'LMQN', ObsOpNum, ObsOpSmoothing

          open(unit=84, file=costFileName, status='unknown', form='formatted')

      elseif (OptimizerNum .eq. 2) then ! LMBM

          write(costFileName,80) 'LMBM', ObsOpNum, ObsOpSmoothing
          write(rmseFileName,81) 'LMBM', ObsOpNum, ObsOpSmoothing

          open(unit=84, file=costFileName, status='unknown', form='formatted')

      elseif (OptimizerNum .eq. 3) then ! DESCON

          write(costFileName,80) 'CGDC', ObsOpNum, ObsOpSmoothing
          write(rmseFileName,81) 'CGDC', ObsOpNum, ObsOpSmoothing

          open(unit=84, file=costFileName, status='unknown', form='formatted')

      elseif (OptimizerNum .eq. 4) then ! CG

          write(costFileName,80) 'CGDT', ObsOpNum, ObsOpSmoothing
          write(rmseFileName,81) 'CGDT', ObsOpNum, ObsOpSmoothing

          open(unit=84, file=costFileName, status='unknown', form='formatted')

      else
          print *,'Unknown OptimizerNum',OptimizerNum
          stop
      endif

      open(unit=86, file=rmseFileName, status='unknown', form='formatted')

      do k=1,nt
          k_start = k

          ! snag the old boundary conditions
          utmp = u(:,:,k)
          vtmp = v(:,:,k)
          phitmp = phi(:,:,k)

          ! run the data assimilation
          if (OptimizerNum .eq. 1) then ! L-BFGS

              INCLUDE 'opt-lbfgs-3dvar.incl'

          elseif (OptimizerNum .eq. 2) then ! LMBM

              INCLUDE 'opt-lmbm-3dvar.incl'

          elseif (OptimizerNum .eq. 3) then ! DESCON

              INCLUDE 'opt-descon-3dvar.incl'

          elseif (OptimizerNum .eq. 4) then ! CG-Descent

              INCLUDE 'opt-cg_descent-3dvar.incl'

          endif

          ! set the new boundary conditions
          call set_boundary_conditions(k, m, n, u(:,:,k), v(:,:,k), phi(:,:,k))

          if (k .gt. 1) then
              ! smooth out the boundary conditions
              utmp = 0.5*(utmp + u(:,:,k))
              vtmp = 0.5*(vtmp + v(:,:,k))
              phitmp = 0.5*(phitmp + phi(:,:,k))

              utmp(2:n-1,2:m-1)   = u(2:n-1,2:m-1,k)
              vtmp(2:n-1,2:m-1)   = v(2:n-1,2:m-1,k)
              phitmp(2:n-1,2:m-1) = phi(2:n-1,2:m-1,k)

              u(:,:,k)   = utmp
              v(:,:,k)   = vtmp
              phi(:,:,k) = phitmp
          endif

          call calculateRmse(k)

          write (86,*) k, sqrt(urmse)/rmseroot, sqrt(vrmse)/rmseroot, sqrt(phirmse)/rmseroot

          if (k .eq. 1) then
              kpass = k
          else
              kpass = k - 1
          endif

          if (k .lt. nt) then

              call calculateRmse(k+1)

              call swe_model(  u(:,:,kpass:k+1), v(:,:,kpass:k+1), &
                             phi(:,:,kpass:k+1), n, m, nt, kpass, k+1, k+1)

              call calculateRmse(k+1)

              do i=2,n-1
                  do j=2,m-1
                      uback(i-1,j-1) = u(i,j,k+1)
                      vback(i-1,j-1) = v(i,j,k+1)
                      phiback(i-1,j-1) = phi(i,j,k+1)
                  enddo
              enddo
           endif

      enddo

      close(unit=84)
      close(unit=86)

      print *, '----------------------------------------------------------------------------------------------------------'

      print *, 'Final RMSE is: '

      open(unit=85, file='hist.dat', status='unknown', form='formatted')

      do k=1,nt
          call calculateRmse(k)
      enddo

      open(unit=85, file='hist.dat', status='unknown', form='formatted')
      write(85,*) (((u(i,j,k),i=1,n),j=1,m),k=1,nt), &
                  (((v(i,j,k),i=1,n),j=1,m),k=1,nt), &
                  (((phi(i,j,k),i=1,n),j=1,m),k=1,nt)
      close(85)

      call finalize()
      deallocate(diag)
      deallocate(ss)
      deallocate(yy)
      deallocate(ww)
      deallocate(w)
      deallocate(grad)
      deallocate(z)
      deallocate(x)
      deallocate(utmp)
      deallocate(vtmp)
      deallocate(phitmp)
      deallocate(dtemp)
      deallocate(xtemp)
      deallocate(gtemp)

      stop
      end
