      program swefourdvar

      use shared
      implicit none

! L-BFGS variables
      integer :: ntrav, msave, nxm, nwork

      common /va15dd/ mp, lp, gtol

      integer :: mp, lp
      double precision :: gtol, oldcost

      double precision, allocatable, dimension(:) :: diag, ss, yy, ww
      double precision :: dzs(10), apd

      integer :: iprint(2), point
      logical :: diagco

! LMBM variables
      integer :: na, mcu, nw
      integer :: ipar(7),iout(3)

      double precision, allocatable, dimension(:) :: w
      double precision :: rpar(8)
      double precision :: time,rtim(2)
      integer :: mc, mcinit

! DESCON variables
      integer nexp, iter, irs, fgcnt, lscnt, ibeta, itheta
      integer itert, irstot, fgtot,lstot, stoptest
      double precision gnorm, epsm
      double precision dc, cc

      logical angle, powell

! CG-Descent variables
      double precision, allocatable, dimension(:) :: dtemp, xtemp, gtemp
      integer :: fcnt, gcnt
      external :: funder_costonly, funder_gradonly

! local variables
      double precision, allocatable, dimension(:) :: grad, z, x
      double precision :: costval, costu, costv, costp, ensfactor
      double precision :: eps

      integer :: i, j, k, icall, iflag, ifunder, ierr

      character(len=64) :: costFileName, rmseFileName

      call initialize()
      call readFiles()

! initialize L-BFGS variables
      ntrav = 25*nctr
      msave = 9
      nxm = msave*nctr
      nwork = nctr+2*msave

      allocate (diag(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for diag"

      allocate (ss(nxm), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for ss"

      allocate (yy(nxm), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for yy"

      allocate (ww(nwork), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for ww"

! initialize LMBM variables
      na = 4
      mcu = 9
      nw = 1 + 9*nctr + 2*nctr*na + 3*na + 2*nctr*(mcu+1) + 3*(mcu+2)*(mcu+1)/2 + 9*(mcu+1)

      allocate (w(nw), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for w"

! initialize CG-descent variables
      allocate (dtemp(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for dtemp"

      allocate (xtemp(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for xtemp"

      allocate (gtemp(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for gtemp"

! initialize local variables
      allocate (grad(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for grad"

      allocate (z(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for z"

      allocate (x(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for x"

!      print*,'Now conducting alpha test of gradient correctness'
      call alphaTest(1.d0)

      u = 0.
      v = 0.
      phi = 0.

      do i=2,n-1
          do j=2,m-1
              u(i,j,1) = uback(i-1,j-1)
              v(i,j,1) = vback(i-1,j-1)
              phi(i,j,1) = phiback(i-1,j-1)
          enddo
      enddo

      call swe_model(u(:,:,1:nt), v(:,:,1:nt), phi(:,:,1:nt), n, m, nt, 1, 1, nt)
      call cost_4dvar(u, v, phi, n, m, nt, uobs, vobs, phiobs, costu, costv, costp, &
          UObsStddev, VObsStddev, PhiObsStddev)
      costval = costu + costv + costp

      print *,'Initial cost is:',costval
      print *,'Initial RMSE is:'

      do k=1,nt
         call calculateRmse(k)
      enddo

      print *, 'Now running test case with: (ObsOpNum,OptimizerNum,ObsOpSmoothing):', ObsOpNum, OptimizerNum, ObsOpSmoothing
      print *, '----------------------------------------------------------------------------------------------------------'

      ! set variables used in common
      z = 0

 80   format('Cost-4DVar-',a4,'-Obsop',i1,'-Smoothing',i1,'.txt')
 81   format('RMSE-4DVar-',a4,'-Obsop',i1,'-Smoothing',i1,'.txt')

      if (OptimizerNum .eq. 1) then ! L-BFGS

          write(costFileName,80) 'LMQN', ObsOpNum, ObsOpSmoothing
          write(rmseFileName,81) 'LMQN', ObsOpNum, ObsOpSmoothing

	      open(unit=84, file=costFileName, status='unknown', form='formatted')

          INCLUDE 'opt-lbfgs-4dvar.incl'

      elseif (OptimizerNum .eq. 2) then ! LMBM

          write(costFileName,80) 'LMBM', ObsOpNum, ObsOpSmoothing
          write(rmseFileName,81) 'LMBM', ObsOpNum, ObsOpSmoothing

          open(unit=84, file=costFileName, status='unknown', form='formatted')

          INCLUDE 'opt-lmbm-4dvar.incl'

      elseif (OptimizerNum .eq. 3) then ! DESCON

          write(costFileName,80) 'CGDC', ObsOpNum, ObsOpSmoothing
          write(rmseFileName,81) 'CGDC', ObsOpNum, ObsOpSmoothing

          open(unit=84, file=costFileName, status='unknown', form='formatted')

          INCLUDE 'opt-descon-4dvar.incl'

      elseif (OptimizerNum .eq. 4) then ! CG-Descent

          write(costFileName,80) 'CGDT', ObsOpNum, ObsOpSmoothing
          write(rmseFileName,81) 'CGDT', ObsOpNum, ObsOpSmoothing

          open(unit=84, file=costFileName, status='unknown', form='formatted')

          INCLUDE 'opt-cg_descent-4dvar.incl'

      else
          print *,'Unknown OptimizerNum',OptimizerNum
          stop
      endif

      close(unit=84)


      print *, '----------------------------------------------------------------------------------------------------------'

      x(1:ns) = MATMUL(sqrtPfU,z(1:ns)) + reshape(uback,(/ns/))
      x(ns+1:2*ns) = MATMUL(sqrtPfV,z(ns+1:2*ns)) + reshape(vback,(/ns/))
      x(2*ns+1:3*ns) = MATMUL(sqrtPfPhi,z(2*ns+1:3*ns)) + &
                       reshape(phiback,(/ns/))

      call convert_from_state(u, v, phi, x, n, m, nt, 1)
      call swe_model(u(:,:,1:nt), v(:,:,1:nt), phi(:,:,1:nt), n, m, nt, 1, 1, nt)

      print *, 'Final cost is: ',costval
      print *, 'Final RMSE is: '

      open(unit=84, file=rmseFileName, status='unknown', form='formatted')

      do k=1,nt
          call calculateRmse(k)

          write (84,*) k, sqrt(urmse)/rmseroot, sqrt(vrmse)/rmseroot, sqrt(phirmse)/rmseroot
      enddo

      close(84)

      open(unit=85, file='hist.dat', status='unknown', form='formatted')
      write(85,*) (((u(i,j,k),i=1,n),j=1,m),k=1,nt), &
                  (((v(i,j,k),i=1,n),j=1,m),k=1,nt), &
                  (((phi(i,j,k),i=1,n),j=1,m),k=1,nt)
      close(85)

! deallocate variables
      call finalize()
      deallocate(diag)
      deallocate(ss)
      deallocate(yy)
      deallocate(ww)
      deallocate(w)
      deallocate(grad)
      deallocate(z)
      deallocate(x)
      deallocate(dtemp)
      deallocate(xtemp)
      deallocate(gtemp)

      stop
      end
